#include <sstream>
#include <algorithm>
#include <cmath>

#include <FL/Enumerations.H>
#include <FL/Fl_Text_Buffer.H>
#include <FL/Fl_Select_Browser.H>
#include <FL/Fl_File_Chooser.H>

#include <FL/Fl_PNG_Image.H>
#include <FL/Fl_JPEG_Image.H>

#include "ShaderViewport.h"
#include "guiBuilder.h"
#include "magickImage.h"

#define	PIOVER180		0.017453292519943295
#define	PI				3.14159265635

#define	MAX_LATITUDE	179.9*PIOVER180
#define MIN_LATITUDE	0.01

#define	MIN_RADIUS		0.01
#define MAX_RADIUS		2000


ShaderViewport::ShaderViewport(int x, int y, int w, int h, const char *label)
	: Fl_Gl_Window(x, y, w, h, label)
	, longitude(45*PIOVER180)
	, latitude(45*PIOVER180)
	, radius(500.0f)
	, firstInit(false)
	, progValid(false)
{
	mode(FL_RGB | FL_DOUBLE | FL_DEPTH);
}

ShaderViewport::~ShaderViewport()
{
	// delete all records
	for(int i = builder->vertexShaderList->size(); i > 0; --i)
	{
		delete ((ShaderRecord*) builder->vertexShaderList->data(i));
	}

	for(int i = builder->fragmentShaderList->size(); i > 0; --i)
	{
		delete ((ShaderRecord*) builder->fragmentShaderList->data(i));
	}

	clearUniforms();

	for(int i = builder->lightBrowser->size(); i > 0; --i)
	{
		delete ((LightRecord*) builder->lightBrowser->data(i));
	}

	for(int i = builder->textureBrowser->size(); i > 0; --i)
	{
		TextureRecord *r = (TextureRecord *) builder->textureBrowser->data(i);

		glActiveTexture(r->glIndex);
		glBindTexture(GL_TEXTURE_2D, 0);
		glDeleteTextures(1, &r->texName);

		delete r->im;
		delete r;
	}
}

void ShaderViewport::draw()
{
	if(!valid())
	{
		if(!firstInit)
		{
			firstInit = true;
			glInit();
		}

		glViewport(0, 0, w(), h());
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		gluPerspective(45, double(w())/h(), 1, 1000);

		glMatrixMode(GL_MODELVIEW);
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glUseProgram(0);

	float co = cos(longitude);
	float so = sin(longitude);

	float ca = cos(latitude);
	float sa = sin(latitude);

	gluLookAt( radius*sa*co, radius*ca, radius*sa*so 
			 , 0, 0, 0
			 , 0, 1, 0
			);

	// do we need to draw the lights?
	int nl = builder->lightBrowser->size();

	for(int i = 1; i <= nl; ++i)
	{
		LightRecord *r = (LightRecord *) builder->lightBrowser->data(i);

		// set the light position and spot direction
		glLightfv(r->glIndex, GL_POSITION, r->pos);
		glLightfv(r->glIndex, GL_SPOT_DIRECTION, r->spotDir);

		// do we need to draw the light?
		if(r->shown)
		{
			drawLight(r->pos);
		}
	}

	// what textures to enable?
	int nt = builder->textureBrowser->size();

	for(int i = 1; i <= nt; ++i)
	{
		TextureRecord *r = (TextureRecord *) builder->textureBrowser->data(i);
		glActiveTexture(r->glIndex);

		if(r->enabled)
		{
			glBindTexture(GL_TEXTURE_2D, r->texName);
			glEnable(GL_TEXTURE_2D);
		}
		else
		{
			glDisable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, 0);
		}
	}

	bool useProg = builder->useProgButton->value();

	if(progValid && useProg)	glUseProgram(program);

	if(builder->drawSphereButton->value())
	{
		double scale = builder->radiusSpinner->value();

		glScaled(scale, scale, scale);
		drawSphere();
	}
	if(builder->drawPlaneButton->value())
	{
		double width = builder->widthSpinner->value();
		double height = builder->heightSpinner->value();

		glScaled(width, height, 1.0);
		drawPlane();
	}
}

int ShaderViewport::handle(int event)
{
	int retVal = 1;

	switch(event)
	{
		case FL_PUSH:
			onMouseDown();
			break;

		case FL_DRAG:
			onMouseDrag();
			break;

		case FL_RELEASE:
			onMouseRelease();
			break;

		case FL_MOUSEWHEEL:
			onMouseWheel();
			break;

		default:
			retVal = Fl_Gl_Window::handle(event);
	}

	return retVal;
}

void ShaderViewport::onMouseDown()
{
	lastX = Fl::event_x();
	lastY = Fl::event_y();
}

void ShaderViewport::onMouseDrag()
{
	int dx = Fl::event_x() - lastX;
	int dy = Fl::event_y() - lastY;

	lastX = Fl::event_x();
	lastY = Fl::event_y();

	const double factor = 0.5*PIOVER180;

	longitude += dx*factor;
	latitude -= dy*factor;

	latitude = latitude > MAX_LATITUDE ? MAX_LATITUDE : latitude;
	latitude = latitude < MIN_LATITUDE ? MIN_LATITUDE : latitude;

	redraw();
}

void ShaderViewport::onMouseRelease()
{
	onMouseDrag();
}

void ShaderViewport::onMouseWheel()
{
	radius *= pow(1.2, Fl::event_dy());

	radius = radius > MAX_RADIUS ? MAX_RADIUS : radius;
	radius = radius < MIN_RADIUS ? MIN_RADIUS : radius;

	redraw();
}


void ShaderViewport::onAddVertexShader()
{
	onAddShader(GL_VERTEX_SHADER, builder->vertexShaderList);
}

void ShaderViewport::onRemoveVertexShader()
{
	onRemoveShader(GL_VERTEX_SHADER, builder->vertexShaderList);
}

void ShaderViewport::onAddFragmentShader()
{
	onAddShader(GL_FRAGMENT_SHADER, builder->fragmentShaderList);
}

void ShaderViewport::onRemoveFragmentShader()
{
	onRemoveShader(GL_FRAGMENT_SHADER, builder->fragmentShaderList);
}

void ShaderViewport::onEditUniform()
{
	int ind = builder->uniformList->value();
	UniformRecord *r = (UniformRecord *) builder->uniformList->data(ind);
	string val = builder->uniformValueInput->value();
	GLint location;

	if(r->size > 1)
	{
		ind = builder->indexSpinner->value();
		r->value[ind] = val;

		std::ostringstream oss;

		oss << r->name << '[' << ind << ']';
		location = glGetUniformLocation(program, oss.str().c_str());
	}
	else
	{
		*r->value = val;
		location = glGetUniformLocation(program, r->name.c_str());
	}

	std::istringstream iss(val);

	float	fvals[16];
	int		ivals[4];
	
	int i = 0;

	switch(r->type)
	{
		case GL_FLOAT_MAT4:	
			iss >> fvals[i++];
			iss >> fvals[i++];
			iss >> fvals[i++];
			iss >> fvals[i++];

		case GL_FLOAT_MAT3x4:
		case GL_FLOAT_MAT4x3:
			iss >> fvals[i++];
			iss >> fvals[i++];
			iss >> fvals[i++];

		case GL_FLOAT_MAT3:
			iss >> fvals[i++];

		case GL_FLOAT_MAT2x4:
		case GL_FLOAT_MAT4x2:
			iss >> fvals[i++];
			iss >> fvals[i++];

		case GL_FLOAT_MAT2x3:
		case GL_FLOAT_MAT3x2:
			iss >> fvals[i++];
			iss >> fvals[i++];

		case GL_FLOAT_MAT2:
		case GL_FLOAT_VEC4:
			iss >> fvals[i++];

		case GL_FLOAT_VEC3:
			iss >> fvals[i++];

		case GL_FLOAT_VEC2:
			iss >> fvals[i++];

		case GL_FLOAT:
			iss >> fvals[i++];
			break;

		case GL_INT_VEC4:
		case GL_BOOL_VEC4:
			iss >> ivals[i++];

		case GL_INT_VEC3:
		case GL_BOOL_VEC3:
			iss >> ivals[i++];

		case GL_INT_VEC2:
		case GL_BOOL_VEC2:
			iss >> ivals[i++];

		case GL_INT:
		case GL_BOOL:
			iss >> ivals[i++];
			break;
	}

	// we've parsed the values, now its time to give them to OpenGL.
	// another ugly switch statement.
	switch(r->type)
	{
		case GL_FLOAT_MAT4:
			glUniformMatrix4fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT3x4:
			glUniformMatrix3x4fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT4x3:
			glUniformMatrix4x3fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT3:
			glUniformMatrix3fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT2x4:
			glUniformMatrix2x4fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT4x2:
			glUniformMatrix4x2fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT2x3:
			glUniformMatrix2x3fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT3x2:
			glUniformMatrix3x2fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_MAT2:
			glUniformMatrix2fv(location, 1, GL_FALSE, fvals);
			break;

		case GL_FLOAT_VEC4:
			glUniform4fv(location, 1, fvals);
			break;

		case GL_FLOAT_VEC3:
			glUniform3fv(location, 1, fvals);
			break;

		case GL_FLOAT_VEC2:
			glUniform2fv(location, 1, fvals);
			break;

		case GL_FLOAT:
			glUniform1fv(location, 1, fvals);
			break;

		case GL_INT_VEC4:
		case GL_BOOL_VEC4:
			glUniform4iv(location, 1, ivals);
			break;

		case GL_INT_VEC3:
		case GL_BOOL_VEC3:
			glUniform3iv(location, 1, ivals);
			break;

		case GL_INT_VEC2:
		case GL_BOOL_VEC2:
			glUniform2iv(location, 1, ivals);
			break;

		case GL_INT:
		case GL_BOOL:
			glUniform1iv(location, 1, ivals);
			break;
	}

	redraw();
}

void ShaderViewport::onUniformIndex()
{
	int ind = builder->uniformList->value();
	UniformRecord *r = (UniformRecord*) builder->uniformList->data(ind);

	ind = builder->indexSpinner->value();

	builder->uniformValueInput->value(r->value[ind].c_str());
}

void ShaderViewport::onSelectUniform()
{
	// do we need to activate the UI?
	int ind = builder->uniformList->value();

	if(ind != 0)
	{
		builder->editUniformButton->activate();
		builder->uniformValueInput->activate();

		UniformRecord *r = (UniformRecord *) builder->uniformList->data(ind);

		if(r->size > 1)
		{
			builder->indexSpinner->activate();
			builder->indexSpinner->range(0, r->size - 1);
			builder->indexSpinner->value(0);
		}
		else
		{
			builder->indexSpinner->deactivate();
		}

		builder->uniformValueInput->value( (*r->value).c_str() );
	}
	else
	{
		builder->uniformValueInput->deactivate();
		builder->indexSpinner->deactivate();
		builder->editUniformButton->deactivate();
	}
}

void ShaderViewport::onLoadUniforms()
{
	updateUniforms();
}

void ShaderViewport::onCompile()
{
	GLint status = GL_TRUE;
	GLint cur;

	builder->outputWin->buffer()->text("Vertex shaders:\n---------------\n");

	// compile the vertex shaders first.
	for(int i = 1; i <= builder->vertexShaderList->size(); ++i)
	{
		ShaderRecord *r = (ShaderRecord *) builder->vertexShaderList->data(i);
		status = status && compileShader(r);
	}

	builder->outputWin->buffer()->append("\nFragment shaders:\n-----------------\n");

	// compile the fragment shaders
	for(int i = 1; i <= builder->fragmentShaderList->size(); ++i)
	{
		ShaderRecord *r = (ShaderRecord*) builder->fragmentShaderList->data(i);
		status = status && compileShader(r);
	}

	// link the program if the above succeeded
	if(status)
	{
		glLinkProgram(program);

		// log the linking of the program
		char buffer[2048];

		glGetProgramInfoLog(program, 2048, 0, buffer);
		builder->outputWin->buffer()->append("\nProgram link:\n-------------\n");
		builder->outputWin->buffer()->append(buffer);

		glGetProgramiv(program, GL_LINK_STATUS, &status);

		if(status)
		{
			builder->outputWin->buffer()->append("\nProgram link successful!\n");
			progValid = true;
		}

//		glValidateProgram(program);
//		glGetProgramiv(program, GL_VALIDATE_STATUS, &status);
//
//		if(status)
//		{
//			builder->outputWin->buffer()->append("Program validated!\n");
//			progValid = true;
//		}
//		else
//		{
//			builder->outputWin->buffer()->append("Program not valid.\n");
//			progValid = false;
//		}
	}
}

void ShaderViewport::onUseProgram()
{
	redraw();
}

void ShaderViewport::setGuiBuilder(GuiBuilder *b)
{
	builder = b;
	b->outputWin->buffer(new Fl_Text_Buffer());
}

void ShaderViewport::glInit()
{
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);

	program = glCreateProgram();

	// fill in the typeMap here.
	typeMap[GL_FLOAT] = "float";
	typeMap[GL_FLOAT_VEC2] = "vec2";
	typeMap[GL_FLOAT_VEC3] = "vec3";
	typeMap[GL_FLOAT_VEC4] = "vec4";
	typeMap[GL_INT] = "int";
	typeMap[GL_INT_VEC2] = "ivec2";
	typeMap[GL_INT_VEC3] = "ivec3";
	typeMap[GL_INT_VEC4] = "ivec4";
	typeMap[GL_BOOL] = "bool";
	typeMap[GL_BOOL_VEC2] = "bvec2";
	typeMap[GL_BOOL_VEC3] = "bvec3";
	typeMap[GL_BOOL_VEC4] = "bvec4";
	typeMap[GL_FLOAT_MAT2] = "mat2";
	typeMap[GL_FLOAT_MAT3] = "mat3";
	typeMap[GL_FLOAT_MAT4] = "mat4";
	typeMap[GL_FLOAT_MAT2x3] = "mat23";
	typeMap[GL_FLOAT_MAT2x4] = "mat24";
	typeMap[GL_FLOAT_MAT3x2] = "mat32";
	typeMap[GL_FLOAT_MAT3x4] = "mat33";
	typeMap[GL_FLOAT_MAT4x2] = "mat42";
	typeMap[GL_FLOAT_MAT4x3] = "mat43";
	typeMap[GL_SAMPLER_1D] = "sampler1D";
	typeMap[GL_SAMPLER_2D] = "sampler2D";
	typeMap[GL_SAMPLER_3D] = "sampler3D";
	typeMap[GL_SAMPLER_CUBE] = "samplerCube";
	typeMap[GL_SAMPLER_1D_SHADOW] = "sampler1DShadow";
	typeMap[GL_SAMPLER_2D_SHADOW] = "sampler2DShadow";

	// fill in the default material data
	updateMaterialData();

	// fill in the points for drawing circles and spheres
	for(int i = 0; i < NPOINTS1; ++i)
	{
		float ang = 2*i*PI/NPOINTS;

		cosines[i] = std::cos(ang);
		sines[i] = std::sin(ang);
	}

	glGetIntegerv(GL_MAX_LIGHTS, &maxLights);
	glGetIntegerv(GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS, &maxTextures);

	GLint maxTexUnits, maxFragUniforms, maxTexCoords, maxVarFloats
		 , maxVertAttribs, maxVertTexUnits, maxVertUniforms;

	glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, &maxTexUnits);
	glGetIntegerv(GL_MAX_TEXTURE_COORDS, &maxTexCoords);

	glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS, &maxFragUniforms);
	glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS, &maxVertUniforms);

	glGetIntegerv(GL_MAX_VARYING_FLOATS, &maxVarFloats);

	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &maxVertAttribs);
	glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, &maxVertTexUnits);


	std::ostringstream oss;
	oss << "Max lights: " << maxLights << std::endl
		<< "Max texture units: " << maxTexUnits << std::endl
		<< "Max combined texture units: " << maxTextures << std::endl
		<< "Max texture coords: " << maxTexCoords << std::endl
		<< "Max vertex texture units: " << maxVertTexUnits << std::endl 
		<< std::endl
		<< "Max vertex uniforms: " << maxVertUniforms << std::endl
		<< "Max fragment uniforms: " << maxFragUniforms << std::endl
		<< "Max vertex attributes: " << maxVertAttribs << std::endl
		<< "Max varying floats: " << maxVarFloats << std::endl;
		
	builder->outputWin->buffer()->append(oss.str().c_str());
}

void ShaderViewport::onAddShader(GLenum type, Fl_Select_Browser *browser)
{
	char *res = fl_file_chooser( "Select shader file"
								 , 0, 0
			);

	if(res)
	{
		std::string fname(res); 

		if(fname.length() > 0)
		{
			ShaderRecord *r = new ShaderRecord;

			r->filename = fname;
			r->number = glCreateShader(type);

			glAttachShader(program, r->number);

			browser->add(fname.c_str(), (void*) r);
		}
	}
}

void ShaderViewport::onRemoveShader(GLenum type, Fl_Select_Browser *browser)
{
	int selection = browser->value();

	if(selection != 0)
	{
		ShaderRecord *r = (ShaderRecord*) browser->data(selection);

		glDetachShader(program, r->number);
		glDeleteShader(r->number);

		delete r;

		browser->remove(selection);
	}
}

void ShaderViewport::updateUniforms()
{
	int num;
	int maxNameLength;

	clearUniforms();

	glGetProgramiv(program, GL_ACTIVE_UNIFORMS, &num);
	glGetProgramiv(program, GL_ACTIVE_UNIFORM_MAX_LENGTH, &maxNameLength);

	std::ostringstream oss;
	oss << "Number of uniforms: " << num << std::endl;

	builder->outputWin->buffer()->text(oss.str().c_str());

	GLchar buffer[512];
	GLint  size;
	GLenum type;

	for(int i = 0; i < num; ++i)
	{
		glGetActiveUniform( program
						  , i
						  , 512
						  , 0
						  , &size
						  , &type
						  , buffer);

		string name = buffer;

		if(name [0] == 'g' && name[1] == 'l' && name[2] == '_')
			continue;

		UniformRecord *r = new UniformRecord;

		r->name = buffer;
		r->type = type;
		r->size = size;
		r->value = new string[r->size];

		for(int i = 0; i < r->size; ++i)	r->value[i] = "0";

		std::ostringstream out;
		out << typeMap[type] << ' ' << r->name;

		if(r->size > 1)
		{
			out << '[' << r->size << ']';
		}

		string label = out.str();

		builder->uniformList->add(label.c_str(), r);
	}
}

void ShaderViewport::clearUniforms()
{
	for(int i = builder->uniformList->size(); i > 0; --i)
	{
		delete [] ((UniformRecord*) builder->uniformList->data(i))->value;
		delete ((UniformRecord*)builder->uniformList->data(i));
	}

	builder->uniformList->clear();
}

GLint ShaderViewport::compileShader(ShaderRecord *r)
{
	Fl_Text_Buffer	buffer;
	buffer.loadfile(r->filename.c_str());

	char *s = buffer.text();
	int l = buffer.length();
	glShaderSource(r->number, 1, (GLchar const **) &s, &l);

	glCompileShader(r->number);

	glGetShaderInfoLog(r->number, l, 0, s);

	builder->outputWin->buffer()->append(s);

	free(s);

	GLint ret;
	glGetShaderiv(r->number, GL_COMPILE_STATUS, &ret);

	return ret;
}

//////////////////////////////
//// Lighting callbacks
//////////////////////////////

void ShaderViewport::onGlobalAmbient()
{
	GLfloat r = builder->gambRed->value();
	GLfloat g = builder->gambGreen->value();
	GLfloat b = builder->gambBlue->value();
	GLfloat a = builder->gambAlpha->value();

	GLfloat arr[] = {r, g, b, a};

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, arr);
	redraw();
}

void ShaderViewport::onLocalViewer()
{
	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER
				, builder->gLocalViewer->value()
					? GL_TRUE
					: GL_FALSE
				);
	redraw();
}

void ShaderViewport::onTwoSided()
{
	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE
				, builder->gTwoSided->value()
					? GL_TRUE
					: GL_FALSE
				);
	redraw();
}

void ShaderViewport::onColorControl()
{
	glLightModeli( GL_LIGHT_MODEL_COLOR_CONTROL
				, builder->gColorControl->value()
					? GL_SEPARATE_SPECULAR_COLOR
					: GL_SINGLE_COLOR
				);
	redraw();
}

void ShaderViewport::onGlobalLighting()
{
	if(builder->gLighting->value())
	{
		glEnable(GL_LIGHTING);
	}
	else
	{
		glDisable(GL_LIGHTING);
	}
	redraw();
}

void ShaderViewport::onAddLight()
{
	int num = findAvailableIndex<LightRecord>( maxLights
											 , builder->lightBrowser
						);

	LightRecord *r = new LightRecord;

	r->index = num;
	r->glIndex = lightGLIndex(num);

	glGetLightfv(r->glIndex, GL_POSITION, r->pos);
	glGetLightfv(r->glIndex, GL_AMBIENT, r->amb);
	glGetLightfv(r->glIndex, GL_DIFFUSE, r->dif);
	glGetLightfv(r->glIndex, GL_SPECULAR, r->spec);
	glGetLightfv(r->glIndex, GL_SPOT_DIRECTION, r->spotDir);
	glGetLightfv(r->glIndex, GL_SPOT_EXPONENT, &r->spotExp);
	glGetLightfv(r->glIndex, GL_SPOT_CUTOFF, &r->spotCutoff);
	glGetLightfv(r->glIndex, GL_CONSTANT_ATTENUATION, &r->attenC);
	glGetLightfv(r->glIndex, GL_LINEAR_ATTENUATION, &r->attenL);
	glGetLightfv(r->glIndex, GL_QUADRATIC_ATTENUATION, &r->attenQ);
	
	r->enabled = false;
	r->shown = false;

	std::ostringstream out;

	out << "LIGHT" << num;
	
	builder->lightBrowser->add(out.str().c_str(), (void *) r);

	if(builder->lightBrowser->size() == maxLights)
	{
		builder->addLightButton->deactivate();
	}
}

void ShaderViewport::onRemoveLight()
{
	int bInd = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(bInd);

	// make sure OpenGL ignores this light
	glDisable(r->glIndex);

	// delete the record and remove the entry.
	builder->lightBrowser->remove(bInd);
	delete r;

	builder->addLightButton->activate();
}

void ShaderViewport::onLightSelection()
{
	int num = builder->lightBrowser->value();

	if(num != 0)
	{
		builder->lightPropGroup->activate();
		updateLightData(num);
	}
	else
	{
		builder->lightPropGroup->deactivate();
	}
}

void ShaderViewport::onLightPosition()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->pos[0] = builder->lightX->value();
	r->pos[1] = builder->lightY->value();
	r->pos[2] = builder->lightZ->value();
	r->pos[3] = builder->lightW->value();

	// position set during draw
	redraw();
}

void ShaderViewport::onLightAmb()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->amb[0] = builder->lightAmbR->value();
	r->amb[1] = builder->lightAmbG->value();
	r->amb[2] = builder->lightAmbB->value();
	r->amb[3] = builder->lightAmbA->value();

	glLightfv(r->glIndex, GL_AMBIENT, r->amb);
	redraw();
}

void ShaderViewport::onLightDiffuse()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->dif[0] = builder->lightDiffuseR->value();
	r->dif[1] = builder->lightDiffuseG->value();
	r->dif[2] = builder->lightDiffuseB->value();
	r->dif[3] = builder->lightDiffuseA->value();

	glLightfv(r->glIndex, GL_DIFFUSE, r->dif);
	redraw();
}

void ShaderViewport::onLightSpec()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->spec[0] = builder->lightSpecR->value();
	r->spec[1] = builder->lightSpecG->value();
	r->spec[2] = builder->lightSpecB->value();
	r->spec[3] = builder->lightSpecA->value();

	glLightfv(r->glIndex, GL_SPECULAR, r->spec);
	redraw();
}

void ShaderViewport::onLightSpotDir()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->spotDir[0] = builder->lightSpotX->value();
	r->spotDir[1] = builder->lightSpotY->value();
	r->spotDir[2] = builder->lightSpotZ->value();

	// spot direction set during draw
	redraw();
}

void ShaderViewport::onLightSpotExponent()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->spotExp = builder->lightSpotExp->value();

	glLightf(r->glIndex, GL_SPOT_EXPONENT, r->spotExp);
	redraw();
}

void ShaderViewport::onLightSpotCutoff()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->spotCutoff = builder->lightSpotCutoff->value();

	glLightf(r->glIndex, GL_SPOT_CUTOFF, r->spotCutoff);
	redraw();
}

void ShaderViewport::onLightAttenConst()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->attenC = builder->lightAttenConst->value();

	glLightf(r->glIndex, GL_CONSTANT_ATTENUATION, r->attenC);
	redraw();
}

void ShaderViewport::onLightAttenLinear()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->attenL = builder->lightAttenLinear->value();

	glLightf(r->glIndex, GL_LINEAR_ATTENUATION, r->attenL);
	redraw();
}

void ShaderViewport::onLightAttenQuad()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->attenQ = builder->lightAttenQuad->value();

	glLightf(r->glIndex, GL_QUADRATIC_ATTENUATION, r->attenQ);
	redraw();
}

void ShaderViewport::onLightEnable()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->enabled = builder->lightEnable->value();

	if(r->enabled)
	{
		glEnable(r->glIndex);
	}
	else
	{
		glDisable(r->glIndex);
	}
	redraw();
}

void ShaderViewport::onLightShow()
{
	int ind = builder->lightBrowser->value();
	LightRecord *r = (LightRecord*) builder->lightBrowser->data(ind);

	r->shown = builder->lightShow->value();

	redraw();
}

template<class T>
int ShaderViewport::findAvailableIndex(GLint max, Fl_Select_Browser *browser)
{
	bool *taken = new bool[max];

	// initialize the array to false
	for(int i = 0; i < max; ++i)	taken[i] = false;

	// find out which lights we actually have
	for(int i = 1; i <= browser->size(); ++i)
	{
		T *r = (T*) browser->data(i);

		taken[r->index] = true;
	}

	int index = 0;

	while(taken[index])	++index;

	delete [] taken;

	return index;
}

void ShaderViewport::updateLightData(int browserIndex)
{
	LightRecord *r = (LightRecord *) builder->lightBrowser->data(browserIndex);

	builder->lightX->value(r->pos[0]);
	builder->lightY->value(r->pos[1]);
	builder->lightZ->value(r->pos[2]);
	builder->lightW->value(r->pos[3]);

	builder->lightAmbR->value(r->amb[0]);
	builder->lightAmbG->value(r->amb[1]);
	builder->lightAmbB->value(r->amb[2]);
	builder->lightAmbA->value(r->amb[3]);

	builder->lightDiffuseR->value(r->dif[0]);
	builder->lightDiffuseG->value(r->dif[1]);
	builder->lightDiffuseB->value(r->dif[2]);
	builder->lightDiffuseA->value(r->dif[3]);

	builder->lightSpecR->value(r->spec[0]);
	builder->lightSpecG->value(r->spec[1]);
	builder->lightSpecB->value(r->spec[2]);
	builder->lightSpecA->value(r->spec[3]);

	builder->lightSpotX->value(r->spotDir[0]);
	builder->lightSpotY->value(r->spotDir[1]);
	builder->lightSpotZ->value(r->spotDir[2]);

	builder->lightSpotExp->value(r->spotExp);
	builder->lightSpotCutoff->value(r->spotCutoff);

	builder->lightAttenConst->value(r->attenC);
	builder->lightAttenLinear->value(r->attenL);
	builder->lightAttenQuad->value(r->attenQ);

	builder->lightEnable->value(r->enabled);
	builder->lightShow->value(r->shown);
}

void ShaderViewport::updateTextureData(int browserIndex)
{
	TextureRecord *r = 
		(TextureRecord*) builder->textureBrowser->data(browserIndex);

	builder->magFilter->value(r->magFilter);
	builder->minFilter->value(r->minFilter);
	builder->texMode->value(r->texMode);
	builder->texInternal->value(r->internalFormat);
	builder->texWrapS->value(r->wrapS);
	builder->texWrapT->value(r->wrapT);

	builder->borderColorR->value(r->borderColor[0]);
	builder->borderColorG->value(r->borderColor[1]);
	builder->borderColorB->value(r->borderColor[2]);
	builder->borderColorA->value(r->borderColor[3]);

	builder->blendColorR->value(r->blendColor[0]);
	builder->blendColorG->value(r->blendColor[1]);
	builder->blendColorB->value(r->blendColor[2]);
	builder->blendColorA->value(r->blendColor[3]);

	builder->texEnabled->value(r->enabled);

	builder->imageBox->setImage(r->im);
}

GLenum ShaderViewport::lightGLIndex(int pos)
{
	GLenum val;

	switch(pos)
	{
		case 0:	val = GL_LIGHT0; break;
		case 1:	val = GL_LIGHT1; break;
		case 2:	val = GL_LIGHT2; break;
		case 3:	val = GL_LIGHT3; break;
		case 4:	val = GL_LIGHT4; break;
		case 5:	val = GL_LIGHT5; break;
		case 6:	val = GL_LIGHT6; break;
		case 7:	val = GL_LIGHT7; break;
		// TODO: what happens when there are more?
	}

	return val;
}

GLenum ShaderViewport::textureGLIndex(int pos)
{
	GLenum val;

	switch(pos)
	{
		case 0: 	val = GL_TEXTURE0; break;
		case 1: 	val = GL_TEXTURE1; break;
		case 2: 	val = GL_TEXTURE2; break;
		case 3: 	val = GL_TEXTURE3; break;
		case 4: 	val = GL_TEXTURE4; break;
		case 5: 	val = GL_TEXTURE5; break;
		case 6: 	val = GL_TEXTURE6; break;
		case 7: 	val = GL_TEXTURE7; break;
		case 8: 	val = GL_TEXTURE8; break;
		case 9: 	val = GL_TEXTURE9; break;
		case 10:	val = GL_TEXTURE10; break;
		case 11:	val = GL_TEXTURE11; break;
		case 12:	val = GL_TEXTURE12; break;
		case 13:	val = GL_TEXTURE13; break;
		case 14:	val = GL_TEXTURE14; break;
		case 15:	val = GL_TEXTURE15; break;
		case 16:	val = GL_TEXTURE16; break;
		case 17:	val = GL_TEXTURE17; break;
		case 18:	val = GL_TEXTURE18; break;
		case 19:	val = GL_TEXTURE19; break;
		case 20:	val = GL_TEXTURE20; break;
		case 21:	val = GL_TEXTURE21; break;
		case 22:	val = GL_TEXTURE22; break;
		case 23:	val = GL_TEXTURE23; break;
		case 24:	val = GL_TEXTURE24; break;
		case 25:	val = GL_TEXTURE25; break;
		case 26:	val = GL_TEXTURE26; break;
		case 27:	val = GL_TEXTURE27; break;
		case 28:	val = GL_TEXTURE28; break;
		case 29:	val = GL_TEXTURE29; break;
		case 30:	val = GL_TEXTURE30; break;
		case 31:	val = GL_TEXTURE31; break;
	}

	return val;
}

void ShaderViewport::drawLight(GLfloat pos[4])
{
	glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_LIGHTING);
//	glDisable(GL_DEPTH_TEST);

	glColor3f(1.0f, 1.0f, 0.0f);

	glPushMatrix();

	// TODO: revisit this to see how to include w coord
	glTranslatef(pos[0], pos[1], pos[2]);
	glScalef(10.0, 10.0, 10.0);

	drawCircle();

	glPushMatrix();
	glRotatef(90, 1, 0, 0);

	drawCircle();
	glPopMatrix();

	glRotatef(90, 0, 1, 0);
	drawCircle();

	glPopMatrix();
	glPopAttrib();
}

void ShaderViewport::drawCircle()
{
	glBegin(GL_LINE_LOOP);

	for(int i = 0; i < NPOINTS; ++i)
	{
		glVertex2f(cosines[i], sines[i]);
	}

	glEnd();
}

void ShaderViewport::drawSphere()
{
//	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	for(int phi = 0; phi < NPOINTS; ++phi)
	{
		float dphi = float(phi)/NPOINTS;
		float dphi1 = float(phi+1)/NPOINTS;

		glBegin(GL_TRIANGLE_STRIP);

		textureCoords(dphi, 0.0f);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glVertex3f(0.0f, 1.0f, 0.0f);

		for(int theta = 1; theta < (NPOINTS >> 1); ++theta)
		{
			float dtheta = float(theta)/(NPOINTS >> 1);

			textureCoords(dphi1, dtheta);
			glNormal3f( sines[theta]*cosines[phi+1]
					  , cosines[theta]
					  , sines[theta]*sines[phi+1]
					);

			glVertex3f( sines[theta]*cosines[phi+1]
					  , cosines[theta]
					  , sines[theta]*sines[phi+1]
					);

			textureCoords(dphi, dtheta);
			glNormal3f( sines[theta]*cosines[phi]
					  , cosines[theta]
					  , sines[theta]*sines[phi]
					);

			glVertex3f( sines[theta]*cosines[phi]
					  , cosines[theta]
					  , sines[theta]*sines[phi]
					);
		}

		textureCoords(dphi, 1.0f);
		glNormal3f(0.0f, -1.0f, 0.0f);
		glVertex3f(0.0f, -1.0f, 0.0f);

		glEnd();
	}
}

void ShaderViewport::drawPlane()
{
	glBegin(GL_QUADS);
		glNormal3f(0, 0, 1);
		textureCoords(0, 1);
		glVertex2f(-1, -1);

		textureCoords(1, 1);
		glVertex2f(1, -1);

		textureCoords(1, 0);
		glVertex2f(1, 1);

		textureCoords(0, 0);
		glVertex2f(-1, 1);
	glEnd();
}

void ShaderViewport::textureCoords(float s, float t)
{
	int nt = builder->textureBrowser->size();

	for(int i = 1; i <= nt; ++i)
	{
		TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(i);

		if(r->enabled)
		{
			glMultiTexCoord2f(r->glIndex, s, t);
		}
	}
//	glTexCoord2f(s, t);
}


//////////////////////////////////////////////////////
// Material Callbacks
//////////////////////////////////////////////////////

void ShaderViewport::onMatAmb()
{
	float vec[] = {
					builder->matAmbR->value()
				  , builder->matAmbG->value()
				  , builder->matAmbB->value()
				  , builder->matAmbA->value()
	};

	glMaterialfv(materialFace(), GL_AMBIENT, vec);
	redraw();
}

void ShaderViewport::onMatDiffuse()
{
	float vec[] = {
					builder->matDiffuseR->value()
				  , builder->matDiffuseG->value()
				  , builder->matDiffuseB->value()
				  , builder->matDiffuseA->value()
	};

	glMaterialfv(materialFace(), GL_DIFFUSE, vec);
	redraw();
}

void ShaderViewport::onMatSpec()
{
	float vec[] = {
					builder->matSpecR->value()
				  , builder->matSpecG->value()
				  , builder->matSpecB->value()
				  , builder->matSpecA->value()
	};

	glMaterialfv(materialFace(), GL_SPECULAR, vec);
	redraw();
}

void ShaderViewport::onMatEmit()
{
	float vec[] = {
					builder->matEmitR->value()
				  , builder->matEmitG->value()
				  , builder->matEmitB->value()
				  , builder->matEmitA->value()
	};

	glMaterialfv(materialFace(), GL_EMISSION, vec);
	redraw();
}

void ShaderViewport::onMatShin()
{
	float val = builder->matShin->value();

	glMaterialf(materialFace(), GL_SHININESS, val);
	redraw();
}

void ShaderViewport::onMatAmbBack()
{
	float vec[] = {
					builder->matAmbRBack->value()
				  , builder->matAmbGBack->value()
				  , builder->matAmbBBack->value()
				  , builder->matAmbABack->value()
	};

	glMaterialfv(GL_BACK, GL_AMBIENT, vec);
	redraw();
}

void ShaderViewport::onMatDiffuseBack()
{
	float vec[] = {
					builder->matDiffuseRBack->value()
				  , builder->matDiffuseGBack->value()
				  , builder->matDiffuseBBack->value()
				  , builder->matDiffuseABack->value()
	};

	glMaterialfv(GL_BACK, GL_DIFFUSE, vec);
	redraw();
}

void ShaderViewport::onMatSpecBack()
{
	float vec[] = {
					builder->matSpecRBack->value()
				  , builder->matSpecGBack->value()
				  , builder->matSpecBBack->value()
				  , builder->matSpecABack->value()
	};

	glMaterialfv(GL_BACK, GL_SPECULAR, vec);
	redraw();
}

void ShaderViewport::onMatEmitBack()
{
	float vec[] = {
					builder->matEmitRBack->value()
				  , builder->matEmitGBack->value()
				  , builder->matEmitBBack->value()
				  , builder->matEmitABack->value()
	};

	glMaterialfv(GL_BACK, GL_EMISSION, vec);
	redraw();
}

void ShaderViewport::onMatShinBack()
{
	float val = builder->matShinBack->value();

	glMaterialf(GL_BACK, GL_SHININESS, val);
	redraw();
}

void ShaderViewport::onSameAsFront()
{
	if(builder->sameAsFront->value())
	{
		builder->backMatGroup->deactivate();

		onMatAmb();
		onMatDiffuse();
		onMatSpec();
		onMatEmit();
		onMatShin();
	}
	else
	{
		builder->backMatGroup->activate();

		onMatAmbBack();
		onMatDiffuseBack();
		onMatSpecBack();
		onMatEmitBack();
		onMatShinBack();
	}
}

void ShaderViewport::onAddTexture()
{
	MagickImage *image = 0;

	// load the image.
	char *res = fl_file_chooser( "Select image file"
								, 0 , 0
			);

	if(res)
	{
		std::string fname = res;

		if(fname.length() > 0)
		{
			// load it
			image = new MagickImage;

			if(!image->loadImage(fname.c_str()))
			{
				delete image;
				image = 0;
			}
		}
	}

	// did we succeed
	if(image)
	{
		// good, create the record
		int num = findAvailableIndex<TextureRecord>( maxTextures
												   , builder->textureBrowser
				);

		TextureRecord *r = new TextureRecord;
		GLenum imval, type;

		r->index = num;
		r->glIndex = textureGLIndex(num);
		r->im = image;
		r->internalFormat = 0;

		// set the pixels
		glActiveTexture(r->glIndex);
		glGenTextures(1, &r->texName);

		glBindTexture(GL_TEXTURE_2D, r->texName);

		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA
					, r->im->width(), r->im->height(), 0
					, GL_RGBA, GL_UNSIGNED_BYTE, r->im->getPixels()
				);

		// query the other values.
		GLint val;
		glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &val);
		glGetTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, r->blendColor);

		switch(val)
		{
			case GL_ADD:		r->texMode = 0;		break;
			case GL_BLEND:		r->texMode = 1;		break;
			case GL_COMBINE:	r->texMode = 2;		break;
			case GL_DECAL:		r->texMode = 3;		break;
			case GL_MODULATE:	r->texMode = 4;		break;
			case GL_REPLACE:	r->texMode = 5;		break;
		}

		glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, &val);
		r->magFilter = val == GL_LINEAR ? 0 : 1;

		glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, &val);
		r->minFilter = val == GL_LINEAR ? 0 : 1;

		glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, &val);

		switch(val)
		{
			case GL_CLAMP:			r->wrapS = 0;	break;
			case GL_CLAMP_TO_EDGE:	r->wrapS = 1;	break;
			case GL_CLAMP_TO_BORDER: r->wrapS = 2;	break;
			case GL_REPEAT:			r->wrapS = 3;	break;
			case GL_MIRRORED_REPEAT: r->wrapS = 4;	break;
		}

		glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, &val);

		switch(val)
		{
			case GL_CLAMP:			r->wrapT = 0;	break;
			case GL_CLAMP_TO_EDGE:	r->wrapT = 1;	break;
			case GL_CLAMP_TO_BORDER: r->wrapT = 2;	break;
			case GL_REPEAT:			r->wrapT = 3;	break;
			case GL_MIRRORED_REPEAT: r->wrapT = 4;	break;
		}

		glGetTexParameterfv( GL_TEXTURE_2D
							, GL_TEXTURE_BORDER_COLOR
							, r->borderColor
				);

		r->enabled = false;

		std::ostringstream oss;
		oss << "GL_TEXTURE" << num;

		builder->textureBrowser->add(oss.str().c_str(), r);

		if(builder->textureBrowser->size() == maxTextures)
		{
			builder->textureAddButton->deactivate();
		}
	}
}

void ShaderViewport::onRemoveTexture()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord *) builder->textureBrowser->data(bInd);

	// make sure OpenGL ignores the texture.
	glActiveTexture(r->glIndex);
	glDisable(GL_TEXTURE_2D);

	// delete the record and remove the entry
	delete r->im;
	delete r;

	builder->textureAddButton->activate();
}

void ShaderViewport::onTextureSelection()
{
	int num = builder->textureBrowser->value();

	if(num != 0)
	{
		builder->texPropGroup->activate();
		updateTextureData(num);
	}
	else
	{
		builder->texPropGroup->deactivate();
		builder->imageBox->setImage(0);
	}

	builder->imageBox->damage(FL_DAMAGE_ALL);
}

void ShaderViewport::onMagFilter()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->magFilter->value();

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexParameteri( GL_TEXTURE_2D
						, GL_TEXTURE_MAG_FILTER
						, v == 0 ? GL_LINEAR : GL_NEAREST
			);

	r->magFilter = v;

	redraw();
}

void ShaderViewport::onMinFilter()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->minFilter->value();

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexParameteri( GL_TEXTURE_2D
						, GL_TEXTURE_MIN_FILTER
						, v == 0 ? GL_LINEAR : GL_NEAREST
			);

	r->minFilter = v;

	redraw();
}

void ShaderViewport::onTextureMode()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->texMode->value();
	GLenum val;

	switch(v)
	{
		case 0:	val = GL_ADD; 		break;
		case 1: val = GL_BLEND;		break;
		case 2: val = GL_COMBINE;	break;
		case 3: val = GL_DECAL;		break;
		case 4: val = GL_MODULATE;	break;
		case 5: val = GL_REPLACE;	break;
	}

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, val);

	r->texMode = v;

	redraw();
}

void ShaderViewport::onInternalFormat()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->texMode->value();
	GLenum val, imval, type;

	switch(v)
	{
		case 0:	val = GL_RGBA;				break;
		case 1: val = GL_RGB;				break;
		case 2: val = GL_LUMINANCE;			break;
		case 3: val = GL_ALPHA;				break;
		case 4: val = GL_LUMINANCE_ALPHA;	break;
		case 5: val = GL_INTENSITY;			break;
	}

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);

	glTexImage2D(GL_TEXTURE_2D, 0, val
				, r->im->width(), r->im->height(), 0
				, GL_RGBA, GL_UNSIGNED_BYTE, r->im->getPixels()
			);

	r->internalFormat = v;

	redraw();
}

void ShaderViewport::onWrapS()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->texWrapS->value();
	GLenum val;

	switch(v)
	{
		case 0:	val = GL_CLAMP;				break;
		case 1: val = GL_CLAMP_TO_EDGE;		break;
		case 2: val = GL_CLAMP_TO_BORDER;	break;
		case 3: val = GL_REPEAT;			break;
		case 4: val = GL_MIRRORED_REPEAT;	break;
	}

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, val);

	r->wrapS = v;

	redraw();
}

void ShaderViewport::onWrapT()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	int v = builder->texWrapT->value();
	GLenum val;

	switch(v)
	{
		case 0:	val = GL_CLAMP;				break;
		case 1: val = GL_CLAMP_TO_EDGE;		break;
		case 2: val = GL_CLAMP_TO_BORDER;	break;
		case 3: val = GL_REPEAT;			break;
		case 4: val = GL_MIRRORED_REPEAT;	break;
	}

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, val);

	r->wrapT = v;

	redraw();
}

void ShaderViewport::onBorderColor()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	r->borderColor[0] = builder->borderColorR->value();
	r->borderColor[1] = builder->borderColorG->value();
	r->borderColor[2] = builder->borderColorB->value();
	r->borderColor[3] = builder->borderColorA->value();

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, r->borderColor);

	redraw();
}

void ShaderViewport::onBlendColor()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	r->blendColor[0] = builder->blendColorR->value();
	r->blendColor[1] = builder->blendColorG->value();
	r->blendColor[2] = builder->blendColorB->value();
	r->blendColor[3] = builder->blendColorA->value();

	glActiveTexture(r->glIndex);
	glBindTexture(GL_TEXTURE_2D, r->texName);
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, r->blendColor);

	redraw();
}

void ShaderViewport::onTexEnabled()
{
	int bInd = builder->textureBrowser->value();
	TextureRecord *r = (TextureRecord*) builder->textureBrowser->data(bInd);

	r->enabled = builder->texEnabled->value();

	redraw();
}

GLenum ShaderViewport::materialFace()
{
	return builder->sameAsFront->value() ? GL_FRONT_AND_BACK : GL_FRONT;
}

void ShaderViewport::updateMaterialData()
{
	float vec[4];

	glGetMaterialfv(GL_FRONT, GL_AMBIENT, vec);

	builder->matAmbR->value(vec[0]);
	builder->matAmbG->value(vec[1]);
	builder->matAmbB->value(vec[2]);
	builder->matAmbA->value(vec[3]);

	glGetMaterialfv(GL_FRONT, GL_DIFFUSE, vec);

	builder->matDiffuseR->value(vec[0]);
	builder->matDiffuseG->value(vec[1]);
	builder->matDiffuseB->value(vec[2]);
	builder->matDiffuseA->value(vec[3]);

	glGetMaterialfv(GL_FRONT, GL_SPECULAR, vec);

	builder->matSpecR->value(vec[0]);
	builder->matSpecG->value(vec[1]);
	builder->matSpecB->value(vec[2]);
	builder->matSpecA->value(vec[3]);

	glGetMaterialfv(GL_FRONT, GL_EMISSION, vec);

	builder->matEmitR->value(vec[0]);
	builder->matEmitG->value(vec[1]);
	builder->matEmitB->value(vec[2]);
	builder->matEmitA->value(vec[3]);

	glGetMaterialfv(GL_FRONT, GL_SHININESS, vec);

	builder->matShin->value(vec[0]);

	glGetMaterialfv(GL_BACK, GL_AMBIENT, vec);

	builder->matAmbRBack->value(vec[0]);
	builder->matAmbGBack->value(vec[1]);
	builder->matAmbBBack->value(vec[2]);
	builder->matAmbABack->value(vec[3]);

	glGetMaterialfv(GL_BACK, GL_DIFFUSE, vec);

	builder->matDiffuseRBack->value(vec[0]);
	builder->matDiffuseGBack->value(vec[1]);
	builder->matDiffuseBBack->value(vec[2]);
	builder->matDiffuseABack->value(vec[3]);

	glGetMaterialfv(GL_BACK, GL_SPECULAR, vec);

	builder->matSpecRBack->value(vec[0]);
	builder->matSpecGBack->value(vec[1]);
	builder->matSpecBBack->value(vec[2]);
	builder->matSpecABack->value(vec[3]);

	glGetMaterialfv(GL_BACK, GL_EMISSION, vec);

	builder->matEmitRBack->value(vec[0]);
	builder->matEmitGBack->value(vec[1]);
	builder->matEmitBBack->value(vec[2]);
	builder->matEmitABack->value(vec[3]);

	glGetMaterialfv(GL_BACK, GL_SHININESS, vec);

	builder->matShinBack->value(vec[0]);
}

void ShaderViewport::onChangeShape()
{
	redraw();
}
